"""
Class: OpenAIAssistantsApiService
Description: Integration with the OpenAI Assistants API.
"""
from openai import AsyncAzureOpenAI
from openai.pagination import AsyncCursorPage
from openai.types import FileObject
from openai.types.beta.threads import Message
from openai.types.beta.threads.message import Attachment
from openai.types.beta.threads.runs import RunStep
from foundationallm.event_handlers import OpenAIAssistantAsyncEventHandler
from foundationallm.operations import OperationsManager
from foundationallm.models.services import OpenAIAssistantsAPIRequest, OpenAIAssistantsAPIResponse
from foundationallm.services import ImageService
from foundationallm.utils import OpenAIAssistantsHelpers

class OpenAIAssistantsApiService:
    """
    Integration with the OpenAI Assistants API.
    """

    def __init__(self, azure_openai_client: AsyncAzureOpenAI, operations_manager: OperationsManager):
        """
        Initializes an OpenAI Assistants API service.

        Parameters
        ----------
        azure_openai_client : AsyncAzureOpenAI
            Async Azure OpenAI client for interacting with the OpenAI Assistants API.
        operations_manager : OperationsManager
            Operations manager for updating the operation.
        """
        self.client = azure_openai_client
        self.operations_manager = operations_manager

    async def add_thread_message_async(self, thread_id: str, role: str, content: str, attachments: list = None):
        return await self.client.beta.threads.messages.create(
            thread_id = thread_id,
            role = role,
            content = content,
            attachments = attachments
        )

    async def run_async(self, request: OpenAIAssistantsAPIRequest, image_service: ImageService) -> OpenAIAssistantsAPIResponse:
        """
        Creates an OpenAI Assistant Run and executes it asynchronously.

        Parameters
        ----------
        request : OpenAIAssistantsAPIRequest
            The request to run with the OpenAI Assistants API service.
        image_service : ImageService
            The image service to use for generating images. If None, no image generation tool will be added.

        Returns
        -------
        OpenAIAssistantsAPIResponse
            The response parsed from the OpenAI Assistants API service response.
        """
        # Process file attachments and assign tools
        attachments = await self._get_request_attachments_async(request)

        # Add User prompt to the thread
        try:
            message = await self.add_thread_message_async(
                thread_id = request.thread_id,
                role = "user",
                content = request.user_prompt,
                attachments = attachments
            )
        except Exception as e:            
            message = f"Error adding user prompt message to thread: {e}"
            return OpenAIAssistantsAPIResponse(
                document_id = request.document_id,
                errors = [ message ]
            )
        
        # Create an image generation tool for the assistant
        if image_service is not None:
            image_generation_tool = {"type": "function", "function": image_service.get_function_definition(function_name='generate_image')}

            # Add the image generation tool to the assistant.
            assistant = await self.client.beta.assistants.retrieve(assistant_id=request.assistant_id)
            tools = assistant.tools

            # If the tools collection already contains the function, remove it
            for tool in tools:
                if tool.type == 'function' and tool.function.name == "generate_image":
                    tools.remove(tool)

            tools.append(image_generation_tool)
            await self.client.beta.assistants.update(assistant_id=request.assistant_id, tools=tools)

        # Create and execute the run
        run = None
        async with self.client.beta.threads.runs.stream(
            thread_id = request.thread_id,
            assistant_id = request.assistant_id,
            event_handler = OpenAIAssistantAsyncEventHandler(self.client, self.operations_manager, request, image_service),
            additional_instructions = "If you generate an image, return the image inline using markdown, along with a detailed description of it.\n\nIMPORTANT: Never display the image more than once in your response!"
        ) as stream:
            await stream.until_done()
            run = await stream.get_final_run()

        #if run.status != "completed":
        #    run = await self.client.beta.threads.runs.retrieve(run_id = run.id, thread_id = request.thread_id)

        if run.status == "failed":
            print(run.last_error.message)
            return OpenAIAssistantsAPIResponse(
                document_id = request.document_id,
                errors = [ run.last_error.message ]
            )
        
        # Retrieve the steps from the run_steps for the analysis
        run_steps = await self.client.beta.threads.runs.steps.list(
          thread_id = request.thread_id,
          run_id = run.id
        )

        analysis_results = await self._parse_run_steps_async(run_steps.data)

        # Retrieve the messages in the thread after the prompt message was appended.
        messages = await self.client.beta.threads.messages.list(
            thread_id=request.thread_id, order="asc", after=message.id
        )

        content = await self._parse_messages_async(messages)

        result = OpenAIAssistantsAPIResponse(
            document_id = request.document_id,
            content = content,
            analysis_results = analysis_results           
        )
        
        if run.usage is not None:            
            result.completion_tokens = run.usage.completion_tokens
            result.prompt_tokens = run.usage.prompt_tokens
            result.total_tokens = run.usage.total_tokens

        return result

    def _create_attachment_from_fileobject(self, file:FileObject):
        """
        Creates an attachment from a file object.

        Parameters
        ----------
        file : FileObject
            The file object to create an attachment from.

        Returns
        -------
        Attachment
            The attachment created from the file object.
        """
        #Get the filename extension if it exists
        filename_extension = file.filename.split('.')[-1] if '.' in file.filename else None
        file_search_supported_extensions = ["c", "cpp", "cs", "css", "doc", "docx", "html", "java", "js", "json", "md", "pdf", "php", "pptx", "py", "rb", "sh", "tex", "ts", "txt"]
        tools = [{"type": "code_interpreter"}]
        if filename_extension in file_search_supported_extensions:
            tools.append({"type": "file_search"})
        return Attachment(
            file_id=file.id,
            tools = tools
        )

    async def _get_request_attachments_async(self, request: OpenAIAssistantsAPIRequest):
        """
        Retrieves the attachments from the request asynchronously.

        Parameters
        ----------
        request : OpenAIAssistantsAPIRequest
            The request to retrieve attachments from.

        Returns
        -------
        List[Attachment]
            The attachments retrieved from the request.
        """
        attachments = []
        if request.attachments:
            for file_id in request.attachments:
                oai_file = await self.client.files.retrieve(file_id)
                attachments.append(
                     self._create_attachment_from_fileobject(oai_file)
                )
        return attachments

    async def _parse_messages_async(self, messages: AsyncCursorPage[Message]):
        """
        Parses the messages from the OpenAI API.

        Parameters
        ----------
        messages : AsyncCursorPage[Message]
            The messages to parse.

        Returns
        -------
        List[MessageContentItemBase]
            The content items within the messages
        """
        ret_content = []
        for msg in messages.data:
            ret_content.extend(OpenAIAssistantsHelpers.parse_message(msg))
        return ret_content

    async def _parse_run_steps_async(self, run_steps: AsyncCursorPage[RunStep]):
        """
        Parses the run steps from the OpenAI API.

        Parameters
        ----------
        run_steps : AsyncCursorPage[RunStep]
            The run steps to parse.

        Returns
        -------
        List[AnalysisResult]
            The analysis results from the run steps.
        """
        analysis_results = []
        for rs in run_steps:
            analysis_result = OpenAIAssistantsHelpers.parse_run_step(rs)
            if analysis_result is not None:
                analysis_results.append(analysis_result)
        return analysis_results



"""
Class: AzureAIAgentServiceWorkflow
Description: Workflow that integrates with the Azure AI Agent Service.
"""
import json
import uuid
from azure.identity.aio import DefaultAzureCredential
from azure.ai.projects.aio import AIProjectClient
from azure.ai.agents.models import (
    AsyncAgentEventHandler,   
    CodeInterpreterToolDefinition,
    FileSearchToolDefinition,
    FunctionTool,
    ListSortOrder,
    MessageAttachment,
    MessageRole, 
    MessageDeltaChunk,
    MessageImageFileContent,
    MessageTextContent,
    MessageTextFilePathAnnotation,
    MessageTextFileCitationAnnotation,    
    RequiredFunctionToolCall,
    RunStep, 
    RunStepCodeInterpreterToolCall,    
    RunStepFunctionToolCall,
    SubmitToolOutputsAction,
    ToolOutput,
    ThreadMessage,    
    ThreadRun
)
from typing import Any,Dict, Optional, List
from logging import Logger
from opentelemetry.trace import Tracer
from opentelemetry.trace import SpanKind

from foundationallm.config import Configuration, UserIdentity
from foundationallm.langchain.common import (
    FoundationaLLMWorkflowBase,
    FoundationaLLMToolBase
)
from foundationallm.models.agents import ExternalAgentWorkflow
from foundationallm.models.constants import (
    AgentCapabilityCategories,
    AzureAIResourceTypeNames,
    ContentArtifactTypeNames,
    ResourceObjectIdPropertyNames,
    ResourceObjectIdPropertyValues,
    ResourceProviderNames,   
    RunnableConfigKeys
)
from foundationallm.models.messages import MessageHistoryItem
from foundationallm.models.orchestration import (
    AnalysisResult,
    CompletionRequestObjectKeys,
    CompletionResponse,
    ContentArtifact,
    FileHistoryItem,
    OpenAITextMessageContentItem,
    OpenAIImageFileMessageContentItem,
    OpenAIFilePathMessageContentItem
)
from foundationallm.operations.operations_manager import OperationsManager
from foundationallm.telemetry import Telemetry

class AzureAIAgentServiceAgentAsyncEventHandler(AsyncAgentEventHandler):
    """
    Event handler for the Azure AI Agent Service.
    """
    def __init__(self,
                    functions: FunctionTool,
                    instance_id: str,
                    operation_id: str,
                    user_prompt:str,
                    user_prompt_rewrite: Optional[str],
                    operations_manager:OperationsManager,
                    project_client: AIProjectClient) -> None:
        super().__init__()
        self.logger : Logger = Telemetry.get_logger("AzureAIAgentServiceAgentAsyncEventHandler")
        self.functions = functions
        self.instance_id = instance_id
        self.operation_id = operation_id
        self.operations_manager = operations_manager
        self.run_steps = {}       
        self.messages = {}
        self.run_id = None
        self.stop_tokens = [".", ",", ":", ";", "\n", " ", ")"] # Use stop tokens to determine when to write to State API.
        self.interim_result = CompletionResponse(
            id = str(uuid.uuid4()),
            operation_id = operation_id,
            user_prompt = user_prompt,
            user_prompt_rewrite = user_prompt_rewrite,
            content = [],
            analysis_results = []
        )

    async def on_message_delta(self, message_delta: MessageDeltaChunk) -> None:       
        if message_delta.text.endswith(tuple(self.stop_tokens)): # Use stop tokens to determine when to write to State API.
            await self.update_state_api_content_async()                  

    async def on_thread_message(self, message: ThreadMessage) -> None:
        if message.status == "created":
            self.messages[message.id] = message            
        if message.status == "completed":
            self.messages[message.id] = message
            await self.update_state_api_content_async()        

    async def on_thread_run(self, run: ThreadRun) -> None:        
        self.run_id = run.id
        if run.status == "failed":
            self.logger.warn(f"Run failed. Error: {run.last_error}")
            print(f"Run failed. Error: {run.last_error}")

        if run.status == "requires_action" and isinstance(run.required_action, SubmitToolOutputsAction):
            tool_calls = run.required_action.submit_tool_outputs.tool_calls
            tool_outputs = []
            for tool_call in tool_calls:
                if isinstance(tool_call, RequiredFunctionToolCall):
                    try:
                        output = await self.functions.execute(tool_call)
                        tool_outputs.append(
                            ToolOutput(
                                tool_call_id=tool_call.id,
                                output=output,
                            )
                        )
                    except Exception as e:
                        self.logger.warn(f"Error executing function tool_call {tool_call.id}: {e}")
                        print(f"Error executing tool_call {tool_call.id}: {e}")            
            if tool_outputs:
                await self.project_client.agents.submit_tool_outputs_to_stream(
                    thread_id=run.thread_id, run_id=run.id, tool_outputs=tool_outputs, event_handler=self
                )
              
    async def on_run_step(self, step: RunStep) -> None:        
        if step.status == "created":
            self.run_steps[step.id] = step
        elif(step.type == "tool_calls" and step.status == "completed"):            
            self.run_steps[step.id] = step

    async def on_error(self, data: str) -> None:
        self.logger.error(f"Event Hander Error: {data}")
        print(f"Event Hander Error: {data}")
        raise Exception(f"Event Hander Error: {data}")
          
    async def on_unhandled_event(self, event_type: str, event_data: Any) -> None:
        self.logger.warn(f"Unhandled Event Type: {event_type}, Data: {event_data}")
        print(f"Unhandled Event Type: {event_type}, Data: {event_data}")

    def get_run_steps(self) -> List[RunStep]:
        """
        Returns the run steps accumulated during the stream.
        """
        return list(self.run_steps.values())

    def get_last_run_id(self) -> str:
        """
        Returns the last run id associated with the stream.
        """
        return self.run_id

    async def update_state_api_analysis_results_async(self):
        self.interim_result.analysis_results = [] # Clear the analysis results list before adding new results.
        for k, v in self.run_steps.items():
            if not v:
                continue
            analysis_result = parse_run_step(v)
            if analysis_result:
                self.interim_result.analysis_results.append(analysis_result)
        await self.operations_manager.set_operation_result_async(self.operation_id, self.instance_id, self.interim_result)

    async def update_state_api_content_async(self):
        self.interim_result.content = [] # Clear the content list before adding new messages.
        for k, v in self.messages.items():
            content_items = parse_message(v)
            self.interim_result.content.extend(content_items)
        await self.operations_manager.set_operation_result_async(self.operation_id, self.instance_id, self.interim_result)

class AzureAIAgentServiceWorkflow(FoundationaLLMWorkflowBase):
    """
    FoundationaLLM workflow implementing an integration with the Azure AI Agent Service.
    """
    def __init__(self,
                 workflow_config: ExternalAgentWorkflow,
                 objects: Dict,
                 tools: List[FoundationaLLMToolBase],
                 user_identity: UserIdentity,
                 config: Configuration,
                 operations_manager: OperationsManager):
        """
        Initializes the FoundationaLLMWorkflowBase class with the workflow configuration.

        Parameters
        ----------
        workflow_config : ExternalAgentWorkflow
            The workflow assigned to the agent.
        objects : dict
            The exploded objects assigned from the agent.
        tools : List[FoundationaLLMToolBase]
            The tools assigned to the agent.
        user_identity : UserIdentity
            The user identity of the user initiating the request.
        config : Configuration
            The application configuration for FoundationaLLM.
        """
        super().__init__(workflow_config, objects, tools, user_identity, config)
        self.name = workflow_config.name
        self.operations_manager = operations_manager
        self.logger : Logger = Telemetry.get_logger(self.name)
        self.tracer : Tracer = Telemetry.get_tracer(self.name)        
        self.instance_id = self.objects.get(CompletionRequestObjectKeys.INSTANCE_ID)
        # Resolves agent_id, thread_id, and project_connection_string
        self.__resolve_ai_agent_service_resources()
       
    async def invoke_async(self,
        operation_id: str,
        user_prompt:str,
        user_prompt_rewrite: Optional[str],
        message_history: List[MessageHistoryItem],
        file_history: List[FileHistoryItem])-> CompletionResponse:
        """
        Invokes the workflow asynchronously.

        Parameters
        ----------
        operation_id : str
            The unique identifier of the FoundationaLLM operation.
        user_prompt : str
            The user prompt message.
        user_prompt_rewrite : str
            The user prompt rewrite message containing additional context to clarify the user's intent.
        message_history : List[BaseMessage]
            The message history.
        file_history : List[FileHistoryItem]
            The file history.
        """
        llm_prompt = user_prompt_rewrite or user_prompt
        analysis_results = []

        # Check file_history for any items for the current attachments, associate the appropriate tool(s).
        # https://learn.microsoft.com/en-us/azure/ai-services/agents/how-to/tools/code-interpreter?tabs=python&pivots=supported-filetypes
        # https://learn.microsoft.com/en-us/azure/ai-services/agents/how-to/tools/file-search?tabs=python&pivots=supported-filetypes
        code_interpreter_extensions = ['.c', '.cpp', '.csv', 'docx', '.html', '.java', '.json', '.md', '.pdf', '.php', '.pptx', '.py', '.rb' '.tex', '.txt', '.css', '.jpeg', '.jpg', '.js', '.gif', '.png', '.tar', '.ts', '.xlsx', '.xml', '.zip']
        file_search_extensions = ['.c', '.cs', '.cpp', '.doc', '.docx', '.html', '.java', '.json', '.md', '.pdf', '.php', '.pptx', '.py', '.rb', '.tex', '.txt', '.css', '.js', '.sh', '.ts']
        message_file_attachments = [
            MessageAttachment(
                tools=[
                    tool for ext, tool in [
                        (code_interpreter_extensions, CodeInterpreterToolDefinition()),
                        (file_search_extensions, FileSearchToolDefinition())
                    ] if any(file.original_file_name.endswith(e) for e in ext)
                ],
                file_id=file.secondary_provider_object_id
            ) for file in file_history
            if file.current_message_attachment  # Grab only the current message attachments
            and file.secondary_provider == ResourceProviderNames.FOUNDATIONALLM_AZUREAI
        ]
        
        async with DefaultAzureCredential(exclude_environment_credential=True) as credential:
            async with AIProjectClient.from_connection_string(
                credential = credential,
                conn_str = self.project_connection_string
            ) as project_client:               
                # Add current message to the thread.
                message = await project_client.agents.create_message(
                    thread_id = self.thread_id,
                    role = MessageRole.USER,
                    content = llm_prompt,
                    attachments=message_file_attachments
                )                
                event_handler = AzureAIAgentServiceAgentAsyncEventHandler(
                    functions = None,
                    instance_id = self.instance_id,
                    operation_id = operation_id,
                    user_prompt = user_prompt,
                    user_prompt_rewrite = user_prompt_rewrite,
                    operations_manager = self.operations_manager,
                    project_client = project_client
                )
                async with await project_client.agents.create_stream(
                    thread_id = self.thread_id, 
                    agent_id = self.agent_id, 
                    event_handler = event_handler
                ) as stream:
                    await stream.until_done()              

                # Get messages from the thread
                messages = await project_client.agents.list_messages(
                    thread_id = self.thread_id,
                    order = ListSortOrder.ASCENDING,
                    after = message.id
                )
                content = [] 
                for message in messages.data:
                    message_content = parse_message(message)              
                    if message_content:
                        content.extend(message_content)
                        
                # event_handler has an accumulator for completed tool calling run steps. 
                # It also has a property with the run_id for the run associated with the stream.
                run_steps = event_handler.get_run_steps()        
                if(len(run_steps) > 0):
                    for step in run_steps:
                        analysis_result = parse_run_step(step)
                        if analysis_result is not None:
                            analysis_results.append(analysis_result)
                run = await project_client.agents.get_run(
                    thread_id=self.thread_id,
                    run_id=event_handler.get_last_run_id()
                )
                # Get final usage details from the run  
                #  prompt_tokens = run.usage.prompt_tokens
                prompt_tokens = 0
                completion_tokens = 0
                total_tokens = 0                  
                if run.usage:
                    prompt_tokens = run.usage.prompt_tokens
                    completion_tokens = run.usage.completion_tokens
                    total_tokens = run.usage.total_tokens                       
                workflow_execution_content_artifact = self.__create_workflow_execution_content_artifact(
                    llm_prompt, 
                    prompt_tokens, 
                    completion_tokens, 
                    total_tokens)
                retvalue = CompletionResponse(
                    operation_id = operation_id,
                    analysis_results = analysis_results,
                    content = content,
                    content_artifacts = [workflow_execution_content_artifact],
                    user_prompt = llm_prompt,
                    full_prompt = run.instructions,
                    completion_tokens = completion_tokens,
                    prompt_tokens = prompt_tokens,
                    total_tokens = total_tokens,
                    total_cost = 0
                )               
                return retvalue    
    
    def __resolve_ai_agent_service_resources(self):
        """
        Resolves the Azure AI Agent Service resources from the objects dictionary.
        Populates the agent_id, thread_id, and project_connection_string.
        """
        # Populate the agent id and thread id from the objects dictionary
        self.agent_id = self.objects.get(CompletionRequestObjectKeys.AZUREAI_AGENT_ID)
        self.thread_id = self.objects.get(CompletionRequestObjectKeys.AZUREAI_AGENT_THREAD_ID)
        if not self.agent_id:
            raise ValueError('Azure AI Agent Service Agent ID is not set in the objects dictionary.')    
        if not self.thread_id:
            raise ValueError('Azure AI Agent Service Thread ID is not set in the objects dictionary.')
        
        project_object_id = self.workflow_config.get_resource_object_id_properties(
            ResourceProviderNames.FOUNDATIONALLM_AZUREAI,
            AzureAIResourceTypeNames.PROJECTS,
            ResourceObjectIdPropertyNames.OBJECT_ROLE,
            ResourceObjectIdPropertyValues.AI_PROJECT            
        )
        if not project_object_id:
            raise ValueError('Azure AI Project ID is not set.')
        
        # Get the Project object from the objects dictionary keyed by project_object_id
        project_object = self.objects.get(project_object_id.object_id)
        if not project_object:
            raise ValueError('Azure AI Project object not found in objects dictionary.')
        self.project_connection_string = project_object['project_connection_string']
        
    def __create_workflow_execution_content_artifact(
            self,
            original_prompt: str,            
            prompt_tokens: int = 0,
            completion_tokens: int = 0,
            completion_time_seconds: float = 0) -> ContentArtifact:
        """
        Creates a content artifact for the workflow execution.

        Parameters
        ----------
        original_prompt : str
            The original prompt.
        prompt_tokens : int
            The number of prompt tokens.
        completion_tokens : int
            The number of completion tokens.
        completion_time_seconds : float
            The completion time in seconds.
        """
        content_artifact = ContentArtifact(id=self.workflow_config.name)
        content_artifact.source = self.workflow_config.name
        content_artifact.type = ContentArtifactTypeNames.WORKFLOW_EXECUTION
        content_artifact.content = original_prompt
        content_artifact.title = self.workflow_config.name
        content_artifact.filepath = None
        content_artifact.metadata = {            
            'prompt_tokens': str(prompt_tokens),
            'completion_tokens': str(completion_tokens),
            'completion_time_seconds': str(completion_time_seconds)
        }
        return content_artifact

# Shared methods
def parse_run_step(run_step: RunStep) -> AnalysisResult:
    """
    Parses a run step from the Azure AI Agent Service.

    Parameters
    ----------
    run_step : RunStep
        The run step to parse.

    Returns
    -------
    AnalysisResult
        The analysis result from the run step.
    OR None
        If the run step does not contain a tool call
        to the code interpreter tool.
    """
    step_details = run_step.step_details
    if step_details and step_details.type == "tool_calls":
        tool_call_detail = step_details.tool_calls
        for details in tool_call_detail:
            if isinstance(details, RunStepCodeInterpreterToolCall):
                result = AnalysisResult(
                    tool_name = details.type,
                    agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                )
                result.tool_input += details.code_interpreter.input  # Source code
                for output in details.code_interpreter.outputs:  # Tool execution output                        
                    if hasattr(output, 'image') and output.image:
                        result.tool_output += "# Generated image file: " + output.image.file_id
                    elif hasattr(output, 'logs') and output.logs:
                        result.tool_output += output.logs
                return result
            elif isinstance(details, RunStepFunctionToolCall):
                result = AnalysisResult(
                    tool_name = details.function.name,
                    agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                )
                result.tool_input += details.function.arguments
                if details.function.output:
                    fn_output = json.loads(details.function.output)
                    if 'data' in fn_output:
                        output_data = json.loads(details.function.output)['data'][0]
                        result.tool_output += json.dumps({"url": output_data['url'], "description": output_data['revised_prompt']})
                    else:
                        # indicative of a failure during the function call, append error message to output
                        print("Error in function call: " + fn_output)
                        result.tool_output += json.dumps(fn_output)                            
                                
    return None
    
def parse_message(message: ThreadMessage):
    """
    Parses a message from the Azure AI Agent Service.

    Parameters
    ----------
    message : Message
        The message to parse.

    Returns
    -------
    List[MessageContentItemBase]
        The content items within the message along with any annotations.
    """
    ret_content = []
    # for each content item in the message
    for ci in message.content:
            match ci:
                case MessageTextContent():
                    text_ci = OpenAITextMessageContentItem(
                        value=ci.text.value,
                        agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                    )
                    for annotation in ci.text.annotations:
                        match annotation:
                            case MessageTextFilePathAnnotation():
                                file_an = OpenAIFilePathMessageContentItem(
                                    file_id=annotation.file_path.file_id,
                                    start_index=annotation.start_index,
                                    end_index=annotation.end_index,
                                    text=annotation.text,
                                    agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                                )
                                text_ci.annotations.append(file_an)
                            case MessageTextFileCitationAnnotation():
                                file_cit = OpenAIFilePathMessageContentItem(
                                    file_id=annotation.file_citation.file_id,
                                    start_index=annotation.start_index,
                                    end_index=annotation.end_index,
                                    text=annotation.text,
                                    agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                                )
                                text_ci.annotations.append(file_cit)
                    ret_content.append(text_ci)
                case MessageImageFileContent():
                    ci_img = OpenAIImageFileMessageContentItem(
                        file_id=ci.image_file.file_id,
                        agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                    )
                    ret_content.append(ci_img)
                #case MessageImageURLContent():
                #    ci_img_url = OpenAIImageFileMessageContentItem(
                #        file_url=ci.image_url.url,
                #        agent_capability_category = AgentCapabilityCategories.AZURE_AI_AGENTS
                #    )
                #    ret_content.append(ci_img_url)
    return ret_content



'''
Manages the plugins in the system.'''
from importlib import import_module, reload
from logging import Logger
import json
import os
import sys

from foundationallm.config import Configuration
from foundationallm.storage import BlobStorageManager

from .external_module import ExternalModule
from .plugin_manager_types import PluginManagerTypes

PLUGIN_MANAGER_CONFIGURATION_NAMESPACE = \
    'FoundationaLLM:APIEndpoints:LangChainAPI:Configuration:ExternalModules'
PLUGIN_MANAGER_STORAGE_ACCOUNT_NAME = \
    f'{PLUGIN_MANAGER_CONFIGURATION_NAMESPACE}:Storage:AccountName'
PLUGIN_MANAGER_STORAGE_AUTHENTICATION_TYPE = \
    f'{PLUGIN_MANAGER_CONFIGURATION_NAMESPACE}:Storage:AuthenticationType'
PLUGIN_MANAGER_STORAGE_CONTAINER = 'resource-provider'
PLUGIN_MANAGER_STORAGE_ROOT_PATH = 'FoundationaLLM.Plugin'
PLUGIN_MANAGER_LOCAL_STORAGE_FOLDER_NAME = \
    'foundationallm_external_modules'

class PluginManager():
    """
    Manages the plugins in the system.
    """

    object_cache : dict[str, object] = {}

    def __init__(self, config:Configuration, logger:Logger):
        """
        Initializes the plugin manager.

        Parameters
        ----------
        config : Configuration
            The configuration object for the system.
        logger : Logger
            The logger object used for logging.
        """
        self.config = config
        self.logger = logger
        self.external_modules: dict[str, ExternalModule] = {}
        self.modules_local_path = f'./{PLUGIN_MANAGER_LOCAL_STORAGE_FOLDER_NAME}'

        if not os.path.exists(self.modules_local_path):
            os.makedirs(self.modules_local_path)

        self.initialized = False
        valid_configuration = False

        try:
            storage_account_name = config.get_value(PLUGIN_MANAGER_STORAGE_ACCOUNT_NAME)
            storage_authentication_type = config.get_value(PLUGIN_MANAGER_STORAGE_AUTHENTICATION_TYPE)
            valid_configuration = True
        except Exception:
            self.logger.exception('The plugin manager configuration is not set up correctly. No plugins will be loaded.')

        if valid_configuration:

            self.logger.info((
                'Initializing plugin manager with the following configuration:\n',
                f'Storage account name:: {storage_account_name}\n',
                f'Storage authentication type: {storage_authentication_type}\n',
                f'Storage container name: {PLUGIN_MANAGER_STORAGE_CONTAINER}\n',
                f'Storage root path: {PLUGIN_MANAGER_STORAGE_ROOT_PATH}\n',
                f'Modules local path: {self.modules_local_path}\n'
            ))

            try:

                self.storage_manager = BlobStorageManager(
                    account_name=storage_account_name,
                    container_name=PLUGIN_MANAGER_STORAGE_CONTAINER,
                    authentication_type=storage_authentication_type
                )

                plugin_blobs = self.storage_manager.list_blobs(
                    f'{PLUGIN_MANAGER_STORAGE_ROOT_PATH}/Python-')
                plugin_blob_names = [blob.name for blob in plugin_blobs]

                for plugin_blob_name in plugin_blob_names:
                    self.logger.info(f'Loading plugin package from: {plugin_blob_name}')

                    plugin_package_content = self.storage_manager.read_file_content(plugin_blob_name)
                    plugin_package = json.loads(plugin_package_content)

                    module_file = plugin_package['package_file_path']
                    module_name = plugin_package['properties']['module_name']
                    plugin_manager_class_names = plugin_package['properties']['plugin_managers']

                    for plugin_manager_class_name in plugin_manager_class_names.split(','):

                        if module_name in self.external_modules:
                            self.external_modules[module_name].plugin_manager_class_names.append(plugin_manager_class_name)
                        else:
                            self.external_modules[module_name] = ExternalModule(
                                module_file=module_file,
                                module_name=module_name,
                                plugin_manager_class_names=[plugin_manager_class_name]
                            )

                self.initialized = True
                self.logger.info('The plugin manager initialized successfully.')

            except Exception as e:
                self.logger.exception('An error occurred while initializing the plugin manager storage manager. No plugins will be loaded.')
                self.logger.error(f'Exception details: {e}')

    def load_external_modules(self, reload_modules:bool=False):
        """
        Loads the external modules into the system.
        """
        if not self.initialized:
            self.logger.error('The plugin manager is not initialized. No plugins will be loaded.')
            return

        loaded_modules = set()

        for module_name in self.external_modules.keys():

            external_module = self.external_modules[module_name]
            module_file_name = external_module.module_file
            local_module_file_name = f'{self.modules_local_path}/{os.path.basename(module_file_name)}'
            self.logger.info(f'Loading module from {module_file_name}')

            try:

                if module_name in loaded_modules:
                    self.logger.info(f'Module {module_name} and all plugin managers are already loaded.')
                    continue
                else:
                    if (self.storage_manager.file_exists(module_file_name)):
                        self.logger.info(f'Copying module file to: {local_module_file_name}')
                        module_file_binary_content = self.storage_manager.read_file_content(module_file_name)
                        with open(local_module_file_name, 'wb') as f:
                            f.write(module_file_binary_content)

                    sys.path.insert(0, local_module_file_name)

                    if reload_modules:
                        external_module.module = reload(external_module.module)
                        self.logger.info(f'Module {module_name} reloaded successfully.')
                    else:
                        external_module.module = import_module(external_module.module_name)
                        self.logger.info(f'Module {module_name} loaded successfully.')

                    external_module.module_loaded = True
                    loaded_modules.add(module_name)

                    for plugin_manager_class_name in external_module.plugin_manager_class_names:
                        # Note the () at the end of the getattr call - this is to call the class constructor, not just get the class.
                        external_module.plugin_managers.append(getattr(external_module.module, plugin_manager_class_name)())

            except Exception as e:
                self.logger.exception(f'An error occurred while loading module: {module_name}')

    def clear_cache(self):
        """
        Clears the object cache.
        """
        self.object_cache.clear()
        self.logger.info('Object cache cleared.')




from pydantic import Field
from typing import Any, Self, Optional, Literal
from foundationallm.langchain.exceptions import LangChainException
from foundationallm.utils import object_utils
from .agent_workflow_base import AgentWorkflowBase

class LangGraphReactAgentWorkflow(AgentWorkflowBase):
    """
    The configuration for a LangGraph ReAct agent workflow.
    """
    type: Literal["langgraph-react-agent-workflow"] = "langgraph-react-agent-workflow"
    graph_recursion_limit: Optional[int] = Field(None, alias="graph_recursion_limit")
   
    @staticmethod
    def from_object(obj: Any) -> Self:

        workflow: LangGraphReactAgentWorkflow = None

        try:
            workflow = LangGraphReactAgentWorkflow(**object_utils.translate_keys(obj))
        except Exception as e:
            raise LangChainException(f"The LangGraph ReAct Agent Workflow object provided is invalid. {str(e)}", 400)
        
        if workflow is None:
            raise LangChainException("The LangGraph ReAct Agent Workflow object provided is invalid.", 400)

        return workflow



"""
Encapsulates knowledge management agent index information.
"""
from typing import Optional
from pydantic import BaseModel
from foundationallm.models.resource_providers.configuration import APIEndpointConfiguration
from foundationallm.models.resource_providers.vectorization import AzureAISearchIndexingProfile

class KnowledgeManagementIndexConfiguration(BaseModel):
    """Knowlege Management Agent metadata model."""
    indexing_profile : AzureAISearchIndexingProfile
    api_endpoint_configuration: APIEndpointConfiguration



class TemplateVariables:
    """The names of template variables."""
    CODE_GENERATION_PROMPT = 'code_generation_prompt'
    CONTEXT = 'context'
    PROMPT = 'prompt'
    CONVERSATION_FILES = 'conversation_files'
    ATTACHED_FILES = 'attached_files'
    CONTEXT_FILES = 'context_files'
    TOOL_RESULTS = 'tool_results'
    FILES_PROMPT = 'files_prompt'



"""
Classes:
    - SettingsBase: Base settings for a vectorization profile
    - ConfigurationReferencesBase: Base class to hold configuration references for a vectorization profile
    - ProfileBase: Base class to hold vectorization profile information
Description: Base class to hold vectorization profile information
"""
from typing import Optional
from foundationallm.models.resource_providers import ResourceBase

class ProfileBase(ResourceBase):
    """
    Base class to hold vectorization profile information.
    """
    settings: Optional[dict] = None
    configuration_references: Optional[dict] = None
